import os
import pickle
from shutil import which

import numpy as np
import pandas as pd
from jinja2 import Environment, FileSystemLoader

EXCLUDED_ATTRS = [
    "locus_seq",
    "guide_long_seq",
    "relative_cut_pos",
    "mmej_patterns",
    "off_targets",
    "_layers",
]


def rev_comp(seq, rna=False):

    complement = {"A": "T", "C": "G", "G": "C", "T": "A", "N": "N"}
    if rna:
        complement = {"A": "T", "C": "G", "G": "C", "T": "A", "U": "A", "-": "-"}
    return "".join(
        [complement[base] if base in complement.keys() else "N" for base in seq[::-1]]
    )


def is_tool(name):
    """Check whether `name` is on PATH and marked as executable."""
    # from whichcraft import which
    return which(name) is not None


def load_cfd_scoring_matrix():
    """Load the cfd scoring matrix."""
    mm_scores = pickle.load(
        open(
            os.path.dirname(os.path.realpath(__file__)) + "/data/mismatch_score.pkl",
            "rb",
        )
    )
    pam_scores = pickle.load(
        open(os.path.dirname(os.path.realpath(__file__)) + "/data/PAM_scores.pkl", "rb")
    )

    return mm_scores, pam_scores


def _guide_instance_to_dict(guide):
    """Convert a class instance to a dictionary."""
    guide_dict = {}
    for attr, value in guide.__dict__.items():
        if attr not in EXCLUDED_ATTRS:
            guide_dict[attr] = value

    for lk, lv in guide.layers.items():
        if isinstance(lv, np.ndarray):
            guide_dict[lk] = lv.mean()
        else:
            guide_dict[lk] = lv

    return guide_dict


def _guides_to_dataframe(guides):
    """Convert a list of guides to a pandas dataframe."""

    guides_ext = [_guide_instance_to_dict(g) for g in guides]
    df = pd.DataFrame(guides_ext)
    df.index = df["id"]

    return df


def _guides_to_csv(guides, file):
    """Write a list of guides to a csv file."""

    guides_df = _guides_to_dataframe(guides)
    guides_df.to_csv(file)

    return None


def _guides_to_bed(guides, file):
    """Write a list of guides to a bed file."""

    guides_df = _guides_to_dataframe(guides).loc[
        :, ["guide_chrom", "guide_start", "guide_end", "id", "rank", "guide_strand"]
    ]
    guides_df.to_csv(file, sep="\t", header=False, index=False)
    return None


def _guides_detailed_table(guides, file):
    """Write a detailed list of guides to a text file."""

    file_loader = FileSystemLoader(
        os.path.dirname(os.path.realpath(__file__)) + "/templates/"
    )
    env = Environment(loader=file_loader)
    env.filters["rev_comp"] = rev_comp
    template = env.get_template("guide_details.j2")
    output_details = template.render(guides=guides)

    with open(file, "w") as f:
        f.write(output_details)


# def render_mmej_table(mmej_patterns):

#     for mmej_pattern in mmej_patterns:
#         left_seq = mmej_pattern["left"][20:]
#         right_seq = mmej_pattern["right"][:-20]
#         pos_mmej = right_seq.index(mmej_pattern["pattern"])
#         right_del = right_seq[:pos_mmej]
#         right_mmej = right_seq[pos_mmej : pos_mmej + len(mmej_pattern["pattern"])]
#         right_rest = right_seq[pos_mmej + len(mmej_pattern["pattern"]) :]
#         mmej_sequence = [
#             left_seq,
#             html.Span([right_del, html.U(right_mmej), right_rest]),
#         ]

#         pattern_layout = html.Div(
#             [
#                 html.Div(
#                     html.Span(mmej_sequence, className="mmej-pattern"),
#                     className="col-7",
#                 ),
#                 html.Div(
#                     html.Span(
#                         round(mmej_pattern["pattern_score"], 2),
#                         className="mmej-pattern",
#                     ),
#                     className="col",
#                 ),
#                 html.Div(
#                     html.Span(mmej_pattern["frame_shift"], className="mmej-pattern"),
#                     className="col",
#                 ),
#                 html.Div(
#                     html.Span(mmej_pattern["deletion_seq"], className="mmej-pattern"),
#                     className="col-2",
#                 ),
#             ],
#             className="row border-bottom border-1",
#         )



from collections import Counter

import numpy as np
from azimuth import model_comparison

from .helpers import load_cfd_scoring_matrix, rev_comp
from .mmej import generate_mmej_patterns
from .off_targets import calculate_cfd_score, calculate_ot_sum_score, run_bowtie


class Guide:
    def __init__(
        self,
        sequence,
        pam_position,
        pam_len,
        strand="+",
        max_flanking_length=75,
        cut_offset=3,
        chromosome="seq",
        start=0,
    ):
        """Guide class.

        Defines a gRNA and its properties.

        Parameters
        ----------
        sequence : str
            gRNA sequence
        pam_position : int
            Position of PAM sequence in gRNA sequence
        pam_len : int
            Length of PAM sequence
        strand : str, optional
            Strand of the gRNA, by default "+"
        max_flanking_length : int, optional
            Max flanking length taken into account for MMEJ calculation, by default 75
        cut_offset : int, optional
            Cut offset from the beginning of PAM, by default 3
        chromosome : str, optional
            Chromosome where the gRNA is located, by default "seq"
        start : int, optional
            Start position of the gRNA on the chromosome, by default 0
        """

        length = len(sequence)
        cut_pos = pam_position - cut_offset

        relative_guide_start = cut_pos - 17
        relative_guide_end = cut_pos + cut_offset + pam_len
        guide_seq = sequence[relative_guide_start:relative_guide_end]
        guide_long_seq = sequence[relative_guide_start - 4 : relative_guide_end + 3]

        if strand == "-":
            cut_pos = pam_position + pam_len + cut_offset
            relative_guide_start = pam_position
            relative_guide_end = cut_pos + 17
            guide_seq = rev_comp(sequence[relative_guide_start:relative_guide_end])
            guide_long_seq = rev_comp(
                sequence[relative_guide_start - 3 : relative_guide_end + 4]
            )

        absolute_cut_pos = start + cut_pos
        absolute_guide_start = start + relative_guide_start
        absolute_guide_end = start + relative_guide_end

        left_slice = cut_pos - max_flanking_length
        right_slice = cut_pos + max_flanking_length
        relative_cut_pos = max_flanking_length

        if left_slice < 0:
            left_slice = 0
            relative_cut_pos = cut_pos
        if right_slice > length:
            right_slice = length

        # Guide properties
        self.locus_seq = sequence[slice(left_slice, right_slice)]
        self.guide_seq = guide_seq
        self.guide_long_seq = guide_long_seq
        self.guide_pam = guide_seq[-pam_len:]
        self.guide_chrom = chromosome
        self.guide_start = absolute_guide_start
        self.guide_end = absolute_guide_end - 1  # 1-based (last position inclusive)
        self.guide_strand = strand
        self.relative_cut_pos = relative_cut_pos
        self.absolute_cut_pos = absolute_cut_pos
        self.rank_score = 0.0
        self.rank = 0
        self.id = ""

        # MMEJ properties
        self.mmej_patterns = []

        # Off-targets
        self.off_targets = []
        self.off_target_str = ""

        # Layers
        self._layers = {}

    def __repr__(self):
        if self.id:
            name = self.id
        else:
            name = "gRNA"
        return f"{name}({self.guide_seq}|{self.guide_chrom}:{self.guide_start}-{self.guide_end}|{self.guide_strand}|)"

    def __getattr__(self, attr):
        return self[attr]

    @property
    def location(self):
        """Returns the location of the guide in the format: chr:start-end.

        Returns
        -------
        str
            String representation of gRNA location
        """
        return f"{self.guide_chrom}:{self.guide_start}-{self.guide_end}"

    def _create_mmej_oof_string(self, mmej_patterns):
        return "|".join([p["frame_shift"] for p in mmej_patterns])

    def simulate_end_joining(self, n_patterns=5, length_weight=20):
        """Simulate Microhomology-Mediated End Joining (MMEJ) events for the
        gRNA.

        MMEJ scoring is based on the Bae et al. 2014 paper (https://doi.org/10.1038/nmeth.3015)

        Parameters
        ----------
        n_patterns : int, optional
            Number of top-scoring MMEJ patterns to keep, by default 5
        length_weight : int, optional
            Lengeth weight, by default 20
        """

        if "N" not in self.locus_seq:
            mmej_patterns = generate_mmej_patterns(
                self.relative_cut_pos, self.locus_seq, length_weight
            )

            if mmej_patterns:
                sorted_mmej_patterns = sorted(
                    mmej_patterns, key=lambda x: -x["pattern_score"]
                )[:n_patterns]

                oof_sum = sum(
                    [
                        p["pattern_score"]
                        for p in sorted_mmej_patterns
                        if p["frame_shift"] == "+"
                    ]
                )

                sum_score = sum([p["pattern_score"] for p in sorted_mmej_patterns])
                oof_score = oof_sum / sum_score * 100
            else:
                sorted_mmej_patterns = None
                sum_score = None
                oof_score = None

            # TODO: refactor
            self.mmej_patterns = sorted_mmej_patterns
            self.mmej_str = self._create_mmej_oof_string(sorted_mmej_patterns)

            self.add_layer("mmej_sum_score", sum_score)
            self.add_layer("mmej_oof_score", oof_score)
        else:
            self.mmej_patterns = None
            self.mmej_str = None

            self.add_layer("mmej_sum_score", 0)
            self.add_layer("mmej_oof_score", 0)

    def find_off_targets(self, genome, **kwargs):
        """Finds off-targets for the guide. The off-targets are found using
        Bowtie. Bowtie index for the genome must be built before running this
        function.

        Notes
        -----
        The off-targets are stored in the `off_targets` attribute. Based on the
        off-targets, the following layers are added to the guide:

        - ot_sum_score: sum of the off-target scores - the lower the better
        - ot_cfd_score_mean: mean of the CFD scores of the off-targets
        - ot_cfd_score_max: max CFD scores of the off-targets
        - ot_cfd_score_sum: sum CFD scores of the off-targets

        Parameters
        ----------
        genome : Genome
            Genome object with the Bowtie index built
        """

        if genome:
            index_path = genome.bowtie_index
        else:
            raise ValueError("No genome / locus specified.")

        if index_path:
            guides_offtargets = run_bowtie(
                guides=[self], genome_index_path=index_path, **kwargs
            )

            self.off_targets = guides_offtargets[0]
            self.add_layer("ot_sum_score", calculate_ot_sum_score(self.off_targets))

            # Load scoring matrices for CFD score calculation
            mm_scores, pam_scores = load_cfd_scoring_matrix()

            # Calculate CFD scores
            cfd_scores = calculate_cfd_score(
                self, self.off_targets, mm_scores, pam_scores
            )

            for ix, cfd in enumerate(cfd_scores.tolist()):
                self.off_targets[ix]["cfd_score"] = cfd

            self.add_layer("ot_cfd_score_mean", cfd_scores.mean())
            self.add_layer("ot_cfd_score_max", cfd_scores.max())
            self.add_layer("ot_cfd_score_sum", cfd_scores.sum())

        else:
            raise ValueError("Bowtie index is not built for the genome / locus.")

    @property
    def off_targets_string(self):
        """Returns a string representation of the off-targets.

        The string representatio captures the number of off-targets with certain number
        of mismatches: n0|n1|n2|n3|n4|n5 (total), where n0 is the number of off-targets
        with 0 mismatches, n1 is the number of off-targets with 1 mismatch, etc.

        For example, if there are 3 off-targets with 0 mismatches, 2 with 1 mismatch,
        1 with 2 mismatches, 0 with 3 mismatches, 5 with 4 mismatches and 1 with 5 mismatches
        the string representation will be "3|2|1|0|5|1 (13)". In the parenthesis, the total
        number of off-targets is given.

        Returns
        -------
        str
            String representation of off-targets.
        """

        # if there are no off-targets, return 0
        if not self.off_targets:
            return "0"

        # count the number of off-targets with certain number of mismatches
        counts = Counter(
            [
                len(ot["mismatches"].keys()) - self.guide_pam.count("N")
                for ot in self.off_targets
            ]
        )

        # create a string representation
        ot_str = "|".join([str(counts[i]) for i in range(max(counts.keys()) + 1)])
        ot_str += f" ({len(self.off_targets)})"

        return ot_str

    @property
    def layers(self):
        return self._layers

    def add_layer(self, name, layer_data):
        """_summary_

        Parameters
        ----------
        name : str
            _description_
        layer_data : float
            _description_
        """
        self._layers[name] = layer_data

    def layer(self, key):
        """_summary_

        Parameters
        ----------
        key : _type_
            _description_

        Returns
        -------
        _type_
            _description_

        Raises
        ------
        ValueError
            _description_
        """
        if key in self._layers.keys():
            return self._layers[key]
        else:
            raise ValueError(f"Layer {key} was not added to the gRNA.")

    def add_azimuth_score(self):
        """Apply Azimuth score to a list of guides.

        Azimuth is a machine learning-based predictive modelling of CRISPR/Cas9 guide efficiency.
        Sometimes its reffered to as Doench 2016 score.

        Described in https://doi.org/10.1038/nbt.3437 (Doench et al., 2016)

        Returns
        -------
        float
            Azimuth score.
        """

        # Azimuth model requires 30bp sequences
        if (
            len(self.guide_long_seq) == 30
            and "N" not in self.guide_long_seq
            and self.guide_pam.endswith("GG")
            and len(self.guide_pam) == 3
        ):
            score_azimuth = float(
                model_comparison.predict(
                    np.array([self.guide_long_seq]), length_audit=True, pam_audit=True
                )[0]
            )
        else:
            score_azimuth = 0.0

        # add layer to gRNA
        self.add_layer("azimuth_score", score_azimuth)

        return score_azimuth



import re

import allel
import mcdm
import numpy as np
import pandas as pd
import pyranges
from pyfaidx import Fasta, Sequence

from .guides import Guide
from .helpers import (
    _guides_detailed_table,
    _guides_to_bed,
    _guides_to_csv,
    _guides_to_dataframe,
    load_cfd_scoring_matrix,
)
from .off_targets import calculate_cfd_score, calculate_ot_sum_score, run_bowtie, get_off_targets_string


class Locus:
    def __init__(
        self,
        sequence,
        name=None,
        start=1,
        end=None,
        genome=None,
        annotation=None,
        **kwargs,
    ):
        """Locus object that represents a genomic locus in which gRNAs are
        searched.

        Parameters
        ----------
        sequence : Sequence, str
            Nucleotide sequence. Can be defined as `pyfaidx.Sequence` or `str`.
        name : str, optional
            Name of the sequence, contig or chromosome, by default None.
        start : int, optional
            Starting position of the locus. It should be 1-based. By default 1.
        end : int, optional
            Ending position of the locus.
        genome : Genome, optional
            Genome object that also includes the genomic locus defined here. It is
            used by default to search for gRNA off-targets.
        annotation : pd.DataFrame, optional
            Annotation table that provides genomic features and is used to annotate
            gRNAs that are found in the locus.

        Examples
        ----------
        >>> import guido
        >>> seq = "TTATCATCCACTCTGACGGGTGGTATTGCGCAACTCCACGCCATCAAACATGTTCAGATTATGCAATCGTGAGTATTCGTTGACCACCGCTTGACCTGTGT"
        >>> loc = guido.Locus(
        ...     sequence=seq, name="AgamP4_2R", start=48714554, end=48714654
        ... )
        >>> loc.find_guides()
        >>> loc.guides
        [gRNA-1(CGCAATACCACCCGTCAGAGTGG|AgamP4_2R:48714561-48714584|-|),
         gRNA-2(TTATCATCCACTCTGACGGGTGG|AgamP4_2R:48714554-48714577|+|),
         gRNA-3(TCTGAACATGTTTGATGGCGTGG|AgamP4_2R:48714589-48714612|-|),
         gRNA-4(CATAATCTGAACATGTTTGATGG|AgamP4_2R:48714594-48714617|-|)]
        """

        self.sequence = sequence
        self.chromosome = name
        self.start = start
        self.intervals = []
        self.genome = genome
        self.annotation = annotation
        self.pam = ""
        self.guides = []
        self._layers = {}

        if isinstance(sequence, Sequence):
            self.sequence = sequence
            self.start = sequence.start
            self.end = sequence.end
            self.chromosome = sequence.name
        else:
            self.sequence = Sequence(
                seq=sequence, start=start, end=end, name=name, **kwargs
            )
            if end and isinstance(end, int):
                self.end = end
            elif not end:
                self.end = self.start + len(self.sequence.seq)

        self.length = len(self.sequence)

    def __repr__(self):
        """Returns a string representation of the locus object."""
        return f"Locus({self.chromosome}:{self.start}-{self.end})"

    def to_dict(self):
        """Converts the locus object to a dictionary."""
        return self.__dict__

    def guide(self, ix):
        """Fetch a guide from the locus by its index or name.

        Parameters
        ----------
        ix : str or int
            Index of the gRNA.

        Returns
        -------
        g: Guide
            Guide object representing a gRNA

        Examples
        --------
        >>> import guido
        >>> seq = "TTATCATCCACTCTGACGGGTGGTATTGCGCAACTCCACGCCATCAAACATGTTCAGATTATGCAATCGTGAGTATTCGTTGACCACCGCTTGACCTGTGT"
        >>> loc = guido.Locus(
        ...     sequence=seq, name="AgamP4_2R", start=48714554, end=48714654
        ... )
        >>> loc.find_guides()
        >>> loc.guide("gRNA-1")
        gRNA-1(CGCAATACCACCCGTCAGAGTGG|AgamP4_2R:48714561-48714584|-|)
        >>> loc.guide(0)
        gRNA-1(CGCAATACCACCCGTCAGAGTGG|AgamP4_2R:48714561-48714584|-|)
        """

        if isinstance(ix, str):
            for g in self.guides:
                if g.id == ix:
                    return g
            else:
                raise ValueError("Provided gRNA index is not valid.")
        elif isinstance(ix, int):
            return self.guides[ix]
        else:
            raise ValueError("Provided gRNA index is not valid.")

    def _flatten_intervals(self, intervals):
        """Flattens overlapping intervals into an union."""
        fi = []
        for start, end in intervals:
            if fi and fi[-1][1] >= start - 1:
                fi[-1][1] = max(fi[-1][1], end)
            else:
                if start < self.start:
                    start = self.start
                if end > self.end:
                    end = self.end
                fi.append([start, end])
        return fi

    def _find_guides_in_interval(self, sequence, start, pam):
        """Finds guides in interval."""
        iupac_dict = {
            "A": ("A", "T"),
            "C": ("C", "G"),
            "G": ("G", "C"),
            "T": ("T", "A"),
            "R": ("[AG]", "[CT]"),
            "Y": ("[CT]", "[AG]"),
            "S": ("[GC]", "[GC]"),
            "W": ("[AT]", "[AT]"),
            "K": ("[GT]", "[AC]"),
            "M": ("[AC]", "[GT]"),
            "B": ("[CGT]", "[ACG]"),
            "D": ("[AGT]", "[ACT]"),
            "H": ("[ACT]", "[AGT]"),
            "V": ("[ACG]", "[CGT]"),
            "N": ("[ACGT]", "[ACGT]"),
        }

        iupac_pam = "".join([iupac_dict[letter][0] for letter in pam])
        rev_iupac_pam = "".join([iupac_dict[letter][1] for letter in pam[::-1]])
        fwd_seq = sequence.upper()

        pams_fw = [m.start() for m in re.finditer(rf"(?=({iupac_pam}))", fwd_seq)]
        pams_rv = [m.start() for m in re.finditer(rf"(?=({rev_iupac_pam}))", fwd_seq)]

        guides_fw = [
            Guide(
                sequence=fwd_seq,
                pam_position=pam_position,
                pam_len=len(pam),
                strand="+",
                chromosome=self.chromosome,
                start=start,
            )
            for pam_position in pams_fw
        ]
        guides_rv = [
            Guide(
                sequence=fwd_seq,
                pam_position=pam_position,
                pam_len=len(pam),
                strand="-",
                chromosome=self.chromosome,
                start=start,
            )
            for pam_position in pams_rv
        ]

        guides = [
            guide for guide in guides_fw + guides_rv if "N" not in guide.guide_seq
        ]
        return guides

    def find_guides(
        self,
        pam="NGG",
        min_flanking_length=0,
        selected_features={"all"},
    ):
        """Find gRNAs in the locus.

        Parameters
        ----------
        pam : str, optional
            gRNA PAM sequence, by default "NGG"
        min_flanking_length : int, optional
            Defines flanking region from the locus where gRNAs are ignored. By default
            0, however `simulate_end_joining()` requires flanking region of 75 bp to
            simulate MMEJ.
        selected_features : str, optional
            Limit gRNA search on only specified genomic features. Features are defined
            in the provided genome annotation file. By default {"all"}

        Returns
        -------
        sorted_guides : list
            List of gRNAs sorted by their position in the locus.

        Examples
        --------
        >>> import guido
        >>> genome = guido.load_genome_from_file(
        ...     guido_file="/Users/nkranjc/imperial/ref/new/AgamP4.guido"
        ... )
        >>> loc = guido.locus_from_coordinates(genome, "AgamP4_2R", 48714541, 48714666)
        >>> loc.find_guides()
        >>> loc.guides
        [gRNA-1(AAGTTTATCATCCACTCTGACGG|AgamP4_2R:48714550-48714572|+|),
        gRNA-2(CGCAATACCACCCGTCAGAGTGG|AgamP4_2R:48714561-48714583|-|),
        gRNA-3(AGTTTATCATCCACTCTGACGGG|AgamP4_2R:48714551-48714573|+|),
        gRNA-4(TTATCATCCACTCTGACGGGTGG|AgamP4_2R:48714554-48714576|+|),
        gRNA-5(TCTGAACATGTTTGATGGCGTGG|AgamP4_2R:48714589-48714611|-|),
        gRNA-6(CATAATCTGAACATGTTTGATGG|AgamP4_2R:48714594-48714616|-|),
        gRNA-7(GTTTAACACAGGTCAAGCGGTGG|AgamP4_2R:48714637-48714659|-|),
        gRNA-8(TATGTTTAACACAGGTCAAGCGG|AgamP4_2R:48714640-48714662|-|)]

        Searching for gRNAs in a specific genomic feature:

        >>> loc.find_guides(selected_features={"exon"})
        >>> loc.guides
        [gRNA-1(AAGTTTATCATCCACTCTGACGG|AgamP4_2R:48714550-48714572|+|),
        gRNA-2(CGCAATACCACCCGTCAGAGTGG|AgamP4_2R:48714561-48714583|-|),
        gRNA-3(AGTTTATCATCCACTCTGACGGG|AgamP4_2R:48714551-48714573|+|),
        gRNA-4(TTATCATCCACTCTGACGGGTGG|AgamP4_2R:48714554-48714576|+|),
        gRNA-5(TCTGAACATGTTTGATGGCGTGG|AgamP4_2R:48714589-48714611|-|),
        gRNA-6(CATAATCTGAACATGTTTGATGG|AgamP4_2R:48714594-48714616|-|)]
        """

        # save searched PAM
        self.pam = pam

        # reset back to empty
        self.guides = []

        # take default locus bounds
        self.intervals = [[self.start, self.end]]

        if "all" not in selected_features and isinstance(self.annotation, pd.DataFrame):
            locus_annotation = self.annotation.query(
                "(Feature == @selected_features) & \
                (Chromosome == @self.chromosome) & \
                (((Start >= @self.start) & (Start <= @self.end)) | \
                ((End >= @self.start) & (End <= @self.end)))"
            ).sort_values("Start")

            # split the locus into smaller loci defined by features
            if len(locus_annotation) > 0:
                self.intervals = self._flatten_intervals(
                    [
                        [f_start, f_end]
                        for f_start, f_end in locus_annotation[["Start", "End"]].values
                    ]
                )

        # search for guides in each locus
        for interval_start, interval_end in self.intervals:

            # check if there is sequence 30 bp upstrean
            rel_interval_start = interval_start - min_flanking_length - self.start
            rel_interval_end = interval_end + min_flanking_length - self.start

            if rel_interval_start < 0:
                rel_interval_start = 0

            if rel_interval_end > len(self.sequence.seq):
                rel_interval_end = len(self.sequence.seq)

            locus_sequence = self.sequence.seq[rel_interval_start:rel_interval_end]
            locus_guides = self._find_guides_in_interval(
                locus_sequence, interval_start - min_flanking_length, pam
            )

            # dont add shorter sequences, check that cut position is in the interval
            locus_guides_filtered = []
            for ix, g in enumerate(locus_guides):
                if (
                    len(g.guide_seq) == 23
                    and interval_start <= g.absolute_cut_pos <= interval_end
                ):
                    locus_guides_filtered.append(g)

            self.guides.extend(locus_guides_filtered)

        # sort guides by cut position
        sorted_guides = []
        for ix, g in enumerate(sorted(self.guides, key=lambda g: g.absolute_cut_pos)):
            g.id = f"gRNA-{ix + 1}"
            sorted_guides.append(g)

        self.guides = sorted_guides

        return sorted_guides

    def simulate_end_joining(self, n_patterns=5, length_weight=20):
        """Simulate end-joining and find MMEJ deletion patterns for each gRNA.

        Microhomology scores are calculated based on proposed scoring model described by
        Bae et al. 2014.

        Parameters
        ----------
        n_patterns : int, optional
            Number of top scored MMEJ deletion patterns reported. By default 5.
        length_weight : int, optional
            Length weight parameter used in MMEJ scoring as defined by Bae et al. 2015.
            By default, 20.
        """

        if len(self.guides) == 0:
            raise ValueError("No gRNAs saved yet.")

        for G in self.guides:
            G.simulate_end_joining(n_patterns, length_weight)

    def find_off_targets(self, external_genome=None, **kwargs):
        """Find off-targets in the genome for each gRNA.

        Parameters
        ----------
        external_genome : Genome, optional
            If provided, off-target search is performed in the external genome rather
            than in the genome which Locus is a part of. By default None.
        """

        # TODO refactor - the same is used in Guide class

        if external_genome:
            index_path = external_genome.bowtie_index
        elif self.genome:
            index_path = self.genome.bowtie_index
        else:
            raise ValueError("No genome / locus specified.")

        if index_path:
            guides_bowtie_offtargets = run_bowtie(
                guides=self.guides, pam=self.pam, genome_index_path=index_path, **kwargs
            )

            mm_scores, pam_scores = load_cfd_scoring_matrix()

            for ix, G in enumerate(self.guides):
                if ix in guides_bowtie_offtargets.keys():
                    G.off_targets = guides_bowtie_offtargets[ix]
                    G.add_layer("ot_sum_score", calculate_ot_sum_score(G.off_targets))
                    G.off_target_str = get_off_targets_string(G.off_targets)

                    # Calculate CFD scores
                    cfd_scores = calculate_cfd_score(
                        G, G.off_targets, mm_scores, pam_scores
                    )

                    for ix, cfd in enumerate(cfd_scores.tolist()):
                        G.off_targets[ix]["cfd_score"] = cfd

                    G.add_layer("ot_cfd_score_mean", cfd_scores.mean())
                    G.add_layer("ot_cfd_score_max", cfd_scores.max())
                    G.add_layer("ot_cfd_score_sum", cfd_scores.sum())

        else:
            raise ValueError("Bowtie index is not built for the genome / locus.")

        return guides_bowtie_offtargets

    def _apply_clipped_layer_data(self, guides, layer_name, layer_data):
        """Apply layer data to guides."""
        if len(guides) > 0:
            for g in self.guides:
                ix = g.guide_start - self.start
                g.add_layer(layer_name, layer_data[ix : ix + 23])

    # TODO: move to util functions
    def _get_guide_regions(self, guide):
        if guide.guide_strand == "-":
            pam_pos = (guide.guide_start, guide.guide_start + 1)
            seed_pos = (guide.guide_start + 3, guide.guide_start + 13)
            seed_small_pos = (guide.guide_start + 3, guide.guide_start + 7)
        else:
            pam_pos = (guide.guide_end - 1, guide.guide_end)
            seed_pos = (guide.guide_end - 13, guide.guide_end - 3)
            seed_small_pos = (guide.guide_end - 7, guide.guide_end - 3)

        guide_pos = (guide.guide_start, guide.guide_end)
        guide_regions = [pam_pos, seed_small_pos, seed_pos, guide_pos]

        return guide_regions

    # TODO: move to util functions
    def _guide_sequence_diversity(self, guide, g, pos):
        """Calculate sequence diversity for each region of the guide."""
        guide_regions = self._get_guide_regions(guide)
        regions_vals = []
        for r in guide_regions:
            try:
                region_loc = pos.locate_range(r[0], r[1])
                region_pos = pos[region_loc]
                region_ac = g[region_loc].count_alleles()
                pi = allel.sequence_diversity(region_pos, region_ac)
                regions_vals.append(pi)
            except Exception:
                regions_vals.append(0)

        return regions_vals

    def _guide_alt_ac(self, guide, g, pos):
        """Calculate alternative allele count for each region of the guide."""
        guide_regions = self._get_guide_regions(guide)
        regions_vals = []
        for r in guide_regions:
            try:
                region_loc = pos.locate_range(r[0], r[1])
                regions_vals.append(g[region_loc].count_alleles()[:, 1:].sum())
            except Exception:
                regions_vals.append(0)

        return regions_vals

    def _guide_n_variants(self, guide, g, pos):
        """Calculate number of variants for each region of the guide."""
        guide_regions = self._get_guide_regions(guide)
        regions_vals = []
        for r in guide_regions:
            try:
                region_loc = pos.locate_range(r[0], r[1])
                regions_vals.append(g[region_loc].n_variants)
            except Exception:
                regions_vals.append(0)

        return regions_vals

    def _apply_variation_layer_data(
        self, guides, layer_name, layer_genotype_data, layer_pos
    ):
        """Apply sequence diversity, alternative allele count and number of
        variants as layers."""
        guide_regions = ["pam", "seed", "small_seed", "guide"]

        if len(guides) > 0:
            for g in self.guides:
                guide_sequence_diversity = self._guide_sequence_diversity(
                    g, layer_genotype_data, layer_pos
                )
                guide_alt_ac = self._guide_alt_ac(g, layer_genotype_data, layer_pos)
                guide_n_variants = self._guide_n_variants(
                    g, layer_genotype_data, layer_pos
                )
                for i in range(len(guide_regions)):
                    g.add_layer(
                        f"var_{layer_name}_{guide_regions[i]}_pi",
                        guide_sequence_diversity[i],
                    )
                    g.add_layer(
                        f"var_{layer_name}_{guide_regions[i]}_alt_ac", guide_alt_ac[i]
                    )
                    g.add_layer(
                        f"var_{layer_name}_{guide_regions[i]}_variants",
                        guide_n_variants[i],
                    )

    @property
    def layers(self):
        """Layers of the locus.

        Returns
        -------
        Layers
            List of layers
        """
        return self._layers

    def add_layer(
        self,
        name,
        layer_data,
        layer_pos=None,
        apply_to_guides=True,
        is_variation=False,
    ):
        """Adds a layer with the data to the locus.

        Parameters
        ----------
        name : str
            Name of the layer
        layer_data : np.ndarray
            Layer data. Needs to be the same shape as the locus.
        apply_to_guides : bool, optional
            Apply layer data to gRNAs when adding it to the locus. By default True.

        Examples
        --------
        >>> locus = Locus("chr1", 100, 200)
        >>> layer_data = np.random.rand(100)
        >>> locus.add_layer("random", layer_data)
        """
        # TODO: make it more intuitive
        self._layers[name] = layer_data

        if apply_to_guides:
            if len(self.guides) > 0:
                if is_variation and layer_pos:
                    self._apply_variation_layer_data(
                        self.guides, name, layer_data, layer_pos
                    )
                else:
                    if layer_data.shape[0] == self.length:
                        self._apply_clipped_layer_data(self.guides, name, layer_data)
                    else:
                        raise ValueError(
                            f"Layer and locus not the same lenght ({layer_data.shape[0]}, {self.length})."
                        )
        else:
            raise ValueError("No guides to apply the data to.")

    def _guide_layers(self):
        """Returns a list of all the layers that are present in the gRNAs."""
        layers = []
        for g in self.guides:
            layers.extend(g.layers.keys())
        return set(layers)

    def _prepare_alt_matrix(self, rank_layer_names, method=np.mean):
        """Prepares numerical matrix with the gRNA layer data to be used later
        in the ranking.

        Parameters
        ----------
        rank_layer_names : list
            List of layer names to be used in the ranking.
        method : [type], optional
            Method to use to combine the layer data, by default np.mean

        Returns
        -------
        np.ndarray
            Matrix with the layer data for each gRNA.
        """
        locus_data = []
        for g in self.guides:
            guide_data = []
            for layer_name in rank_layer_names:
                if (
                    layer_name not in g.layers.keys()
                    and layer_name not in self.layers.keys()
                ):
                    print(g, layer_name, self.layers.keys(), g.layers.keys())
                    raise ValueError(
                        f"Layer {layer_name} does not exist on `Locus` or `Guide` object."
                    )

                if (
                    layer_name not in g.layers.keys()
                    and layer_name in self.layers.keys()
                ):
                    self._apply_clipped_layer_data(
                        [g], layer_name=layer_name, layer_data=self._layers[layer_name]
                    )

                layer_data = g._layers[layer_name]

                if (
                    isinstance(layer_data, np.ndarray)
                    and layer_data.squeeze().ndim == 1
                ):
                    layer_data = method(layer_data)
                elif isinstance(layer_data, float) or isinstance(layer_data, int):
                    layer_data = float(layer_data)
                elif isinstance(layer_data, (pd.Series, list)):
                    layer_data = method(np.array(layer_data))
                else:
                    raise TypeError("Type of layer data is not valid.")

                guide_data.append(layer_data)
            locus_data.append(guide_data)

        return np.array(locus_data)

    def rank_guides(
        self,
        layer_names=None,
        layer_is_benefit=None,
        weight_vector=None,
        ranking_method="TOPSIS",
        norm_method="Vector",
    ):
        """Ranks guides based on the layer data.

        Returns
        -------
        list
            List of ranked guides.
        """

        if len(self.guides) == 0:
            raise ValueError(
                "No gRNAs to rank. Try running `find_guides()` method first."
            )

        # get layer names registered on individual guides
        guide_layer_names = sorted(
            list(
                set(
                    [
                        guide_layer_name
                        for g in self.guides
                        for guide_layer_name in g.layers.keys()
                    ]
                )
            )
        )

        # check if requested layer exists on guides or on locus
        if layer_names:
            for layer in guide_layer_names:
                if layer not in self._layers.keys() and layer not in guide_layer_names:
                    raise ValueError(
                        f"Layer {layer} is not added to `Locus` or `Guide`."
                    )
        else:
            layer_names = list(guide_layer_names) + list(self._layers.keys())

        x_matrix = self._prepare_alt_matrix(rank_layer_names=layer_names)
        rank_scores = mcdm.rank(
            x_matrix,
            n_method=norm_method,
            w_vector=weight_vector,
            is_benefit_x=layer_is_benefit,
            s_method=ranking_method,
            alt_names=[g.id for g in self.guides],
        )

        # order ranks from best to worst
        rank_asc = np.argsort([-r[1] for r in rank_scores])

        for i, (g_id, rank_score) in enumerate(rank_scores):
            self.guide(g_id).rank_score = rank_score
            self.guide(g_id).rank = rank_asc[i] + 1

        return rank_scores

    def guides_to_dataframe(self):
        """Returns gRNAs in Pandas dataframe."""
        return _guides_to_dataframe(self.guides)

    def guides_to_csv(self, filename):
        """Save gRNAs in CSV file."""
        if filename:
            return _guides_to_csv(self.guides, filename)
        else:
            raise ValueError("Filename required to save CSV with gRNAs.")

    # TODO: bed is 0-based, but gRNAs are 1-based
    def guides_to_bed(self, filename):
        """Save gRNAs in BED file."""
        if filename:
            return _guides_to_bed(self.guides, filename)
        else:
            raise ValueError("Filename required to save BED with gRNAs.")

    def guides_detailed_table(self, filename):
        """Save gRNAs in a detailed text file."""
        if filename:
            return _guides_detailed_table(self.guides, filename)
        else:
            raise ValueError("Filename required to save CSV with gRNAs.")

    # TODO: plot the locus

    def add_azimuth_score(self):
        """Apply Azimuth score to a list of guides.

        Azimuth is a machine learning-based predictive modelling of CRISPR/Cas9 guide efficiency. Sometimes its reffered to as
        Doench 2016 score.

        Described in https://doi.org/10.1038/nbt.3437 (Doench et al., 2016)
        """

        return [g.add_azimuth_score() for g in self.guides]


"""
Locus creation ------------------
"""


def _prepare_annotation(annotation_file_abspath, as_df=True):
    """Prepare annotation file for use with pyranges."""
    if annotation_file_abspath.suffix in [".gff3", ".gff"]:
        ann_db = pyranges.read_gff3(str(annotation_file_abspath), as_df=as_df)
        if as_df:
            ann_db = ann_db.rename(columns={"Name": "Exon"})  # type: ignore
    elif annotation_file_abspath.suffix in [".gtf"]:
        ann_db = pyranges.read_gtf(str(annotation_file_abspath), as_df=as_df)
        if as_df:
            ann_db = ann_db.rename(columns={"gene_id": "ID", "exon_number": "Exon"})  # type: ignore
    else:
        raise ValueError(
            "Annotation file not recognised. Annotation file needs to be GFF3 or GTF formnat."
        )
    return ann_db


def locus_from_coordinates(genome, chromosome, start, end):
    """Create a locus from coordinates. Coordinates are 1-based. If annotation
    file is provided, it will be used to annotate the locus.

    Parameters
    ----------
    genome : Genome
        Genome object. Can be created using `Genome` class.
    chromosome : str
        Chromosome name.
    start : int
        Start position.
    end : int
        End position.

    Returns
    -------
    Locus
        Locus object.
    """

    locus_sequence = Fasta(str(genome.genome_file_abspath)).get_seq(
        chromosome, start, end
    )

    if genome.annotation_file_abspath and genome.annotation_file_abspath.exists():
        ann_db = _prepare_annotation(genome.annotation_file_abspath, as_df=False)
        locus_annotation = ann_db.intersect(
            pyranges.PyRanges(chromosomes=[chromosome], starts=[start], ends=[end])
        ).df

        # TODO: rename Name to Exon not necessarily applicable to all annotations
        if genome.annotation_file_abspath.suffix.lower() in [".gff3", ".gff"]:
            locus_annotation = locus_annotation.rename(columns={"Name": "Exon"})  # type: ignore
        elif genome.annotation_file_abspath.suffix.lower() in [".gtf"]:
            locus_annotation = locus_annotation.rename(columns={"gene_id": "ID", "exon_number": "Exon"})  # type: ignore

        if len(locus_annotation) > 0:
            return Locus(
                genome=genome, sequence=locus_sequence, annotation=locus_annotation
            )
        else:
            return Locus(
                genome=genome, sequence=locus_sequence, annotation=None
            )  # TODO: refactor
    else:
        return Locus(genome=genome, sequence=locus_sequence, annotation=None)


def locus_from_sequence(sequence, sequence_name=None):
    """Create a locus from sequence.

    Parameters
    ----------
    sequence : str
        DNA sequence
    sequence_name : str, optional
        Sequence name, by default None

    Returns
    -------
    Locus
        Object representing a locus from given sequence.
    """
    return Locus(sequence=sequence, name=sequence_name)


def locus_from_gene(genome, gene_name):
    """Create a locus from gene name. If annotation file is provided, it will
    be used to annotate the locus.

    Parameters
    ----------
    genome : Genome
        Genome object. Can be created using `Genome` class.
    gene_name : str
        Gene name. Needs to be present in the annotation file.

    Returns
    -------
    Locus
        Locus object.
    """
    if genome.annotation_file_abspath.exists():
        try:
            ann_db = _prepare_annotation(genome.annotation_file_abspath, as_df=True)
            gene_annotation = ann_db.query(
                "ID == @gene_name & (Feature == 'protein_coding_gene' | Feature == 'gene')"
            )

            chromosome = gene_annotation.Chromosome.values[0]
            start = int(gene_annotation.Start) + 1  # pyranges is 0-based
            end = int(gene_annotation.End)

            locus_annotation = ann_db.query(
                f"(ID == @gene_name) & (Chromosome == @chromosome) &  \
                  (((Start >= {start}) & (Start <= {end})) | \
                  ((End >= {start}) & (End <= {end})))"
            )

            locus_sequence = Fasta(
                str(genome.genome_file_abspath), one_based_attributes=True
            ).get_seq(chromosome, start, end)

            return Locus(
                genome=genome, sequence=locus_sequence, annotation=locus_annotation
            )

        except Exception as e:
            raise ValueError(f"Gene {gene_name} not found. (Error: {e})")
    else:
        raise ValueError("Annotation file not valid.")  # TODO: fix message



import itertools
import math
import re

import numpy as np


def generate_mmej_patterns(rel_cut_position, sequence, length_weight):
    """Start with predefined k-mer length and extend it until it finds more
    than one match in sequence."""

    # split the sequence at the cut site
    left_seq = sequence[:rel_cut_position]
    right_seq = sequence[rel_cut_position:]

    # kmers list
    kmers = []
    # combination of positions
    all_combinations = []
    # k-mer starting length
    min_kmer_length = 2

    # expand k-mer length
    for k in list(range(min_kmer_length, len(left_seq)))[::-1]:

        # iterate through sequence
        for i in range(len(left_seq) - k + 1):
            kmer = left_seq[i : i + k]

            # check if the k-mer is in both sequences and
            # hasn't been found before
            if kmer in right_seq and kmer not in kmers:

                # find positions of kmer
                p = re.compile(kmer)

                # find positions of patterns in each sequence
                left_positions = [(m.start(), m.end()) for m in p.finditer(left_seq)]
                right_positions = [(m.start(), m.end()) for m in p.finditer(right_seq)]

                # get combinations
                pos_combinations = [
                    c[0] + c[1]
                    for c in list(itertools.product(left_positions, right_positions))
                ]

                all_combinations.extend(pos_combinations)

                # save kmer
                kmers.append(kmer)

    if all_combinations:

        combinations_array = np.array(all_combinations)
        m1 = combinations_array[:, 0, None] <= combinations_array[:, 0]
        m2 = combinations_array[:, 1, None] >= combinations_array[:, 1]

        m3 = combinations_array[:, 2, None] <= combinations_array[:, 2]
        m4 = combinations_array[:, 3, None] >= combinations_array[:, 3]

        m12 = m1 & m2
        m34 = m3 & m4

        subpattern_rows = np.triu((m12 & m34), 1).any(0)

        # all mmej patterns
        mmej_patterns = []

        # generate microhomology for every combination
        for combination in combinations_array[~subpattern_rows]:

            # save output to dict
            pattern_dict = {}

            left_start, left_end, right_start, right_end = combination
            pattern_seq = left_seq[left_start:left_end]
            pattern_GC = pattern_seq.count("G") + pattern_seq.count("C")

            # deletion length
            left_deletion_length = len(left_seq) - left_start
            right_deletion_length = right_start
            deletion_length = left_deletion_length + right_deletion_length
            deletion_seq = left_seq[left_start:] + right_seq[:right_start]

            # score pattern
            length_factor = round(1 / math.exp((deletion_length) / (length_weight)), 3)
            pattern_score = (
                100
                * length_factor
                * ((len(pattern_seq) - pattern_GC) + (pattern_GC * 2))
            )

            # frame shift
            if deletion_length % 3 == 0:
                frame_shift = "-"
            else:
                frame_shift = "+"

            # create dictionary
            pattern_dict["left"] = left_seq[:left_start] + "-" * left_deletion_length
            pattern_dict["left_seq"] = left_seq[:left_start]
            pattern_dict["left_seq_position"] = left_start
            pattern_dict["right"] = (
                "+" * right_deletion_length + right_seq[right_start:]
            )
            pattern_dict["right_seq"] = right_seq[right_start:]
            pattern_dict["right_seq_position"] = right_start + len(deletion_seq)
            pattern_dict["pattern"] = pattern_seq
            pattern_dict["pattern_len"] = len(pattern_seq)
            pattern_dict["pattern_score"] = pattern_score
            pattern_dict["deletion_seq"] = deletion_seq
            pattern_dict["frame_shift"] = frame_shift

            # add to list
            mmej_patterns.append(pattern_dict)

        return mmej_patterns

    else:
        return None



import pickle
import subprocess
from pathlib import Path

from pyfaidx import Faidx, Fasta


def check_file(filename, supported_ext):
    """Check if file exists and has the right extension.

    Parameters
    ----------
    filename : str
        Path to the file.
    supported_ext : str
        Supported file extensions.

    Returns
    -------
    bool
        True if file exists and has the right extension, False otherwise.
    """
    filename = Path(filename)
    if filename.exists() and filename.suffix in supported_ext:
        return True
    else:
        return False


class Genome:
    def __init__(
        self,
        genome_name,
        genome_file_abspath=None,
        annotation_file_abspath=None,
        bowtie_index_abspath=None,
    ):
        """Genome class.

        Parameters
        ----------
        genome_name : str
            Name of the genome.
        genome_file_abspath : str, optional
            Path to the genome Fasta file, by default None
        annotation_file_abspath : str, optional
            Path to the genome annotation file (GTF or GFF3), by default None
        bowtie_index_abspath : str, optional
            Path to the bowtie index files, by default None
        """
        self.genome_name = genome_name
        self._bowtie_ignore = False
        self.bowtie_index = None

        if genome_file_abspath:
            if check_file(genome_file_abspath, [".fa", ".fasta", ".fna"]):
                self.genome_file_abspath = Path(genome_file_abspath).resolve()
            else:
                raise ValueError(
                    f"Genome file {genome_file_abspath} does noth exist or is not in the right format."
                )

        if bowtie_index_abspath:
            self.bowtie_index = Path(bowtie_index_abspath)
            self._bowtie_ignore = True

        if annotation_file_abspath:
            if check_file(annotation_file_abspath, [".gtf", ".gff3"]):
                self.annotation_file_abspath = Path(annotation_file_abspath).resolve()
            else:
                raise ValueError(
                    f"Annotation file file {genome_file_abspath} does noth exist or is not in the right format."
                )
        else:
            self.annotation_file_abspath = None

    @property
    def is_built(self):
        """Check if genome index is built.

        Returns
        -------
        Bool
            True if genome index is built, False otherwise.
        """
        ready = []
        if self.genome_file_abspath and self.fai_file:
            ready.append(True)
        else:
            return False

        if self.annotation_file_abspath and not self.tbi_file:
            ready.append(False)
        else:
            ready.append(True)

        bowtie_index_files = [
            i
            for i in self.genome_file_abspath.parent.glob("*ebwt")
            if f"{self.genome_name}.rev." in str(i) and f"{self.genome_name}." in str(i)
        ]
        if not self._bowtie_ignore and any(bowtie_index_files):
            ready.append(True)
        elif not self._bowtie_ignore and not any(bowtie_index_files):
            ready.append(False)

        return all(ready)

    def build(
        self,
        n_threads=1,
        save_pickle=True,
        bowtie_ignore=False,
        bowtie_path="",
    ):
        """Build genome index.

        This method creates index files for the genome and annotation files and saves them
        in a `.guido` file in the same directory as the genome Fasta file. This file can be
        later loaded using :meth:`guido.genome.load_genome_from_file` without having to
        re-build the index.

        Parameters
        ----------
        n_threads : int, optional
            Number of threads, by default 1
        save_pickle : bool, optional
            Pickle the dictionary into `.guido` file, by default True
        bowtie_ignore : bool, optional
            Ignore building bowtie index. Use if you already have bowtie index built, by default False
        bowtie_path : str, optional
            Path to bowtie binary if it's not in the path, by default ""
        """

        self._bowtie_ignore = bowtie_ignore
        if self.annotation_file_abspath:
            print("Indexing genome annotation.")
            annotation_sorted_bgz = self.annotation_file_abspath.with_suffix(
                str(self.annotation_file_abspath.suffix) + ".gz"
            )
            annotation_sort_bgz_cmd = f"sort -k1,1 -k4,4n {str(self.annotation_file_abspath)} | bgzip > {annotation_sorted_bgz}"
            self.annotation_sorted_gz_file = annotation_sorted_bgz

            _ = subprocess.run(annotation_sort_bgz_cmd, shell=True)
            tabix_cmd = f"tabix {str(self.annotation_sorted_gz_file)}"

            _ = subprocess.run(tabix_cmd, stderr=subprocess.PIPE, shell=True)
            self.tbi_file = self.annotation_sorted_gz_file.with_suffix(".tbi")

        if self.genome_file_abspath.exists():
            Faidx(str(self.genome_file_abspath))
            self.fai_file = self.genome_file_abspath.with_suffix(
                str(self.genome_file_abspath.suffix) + ".fai"
            )

            if not self._bowtie_ignore:
                bowtie_index_command = f"{bowtie_path}bowtie-build {str(self.genome_file_abspath)} {str(self.genome_file_abspath.parent / self.genome_name)} --threads {n_threads}"
                print("Building Bowtie index")
                bowtie_index_proc = subprocess.run(
                    bowtie_index_command.strip(),
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    shell=True,
                )
                if bowtie_index_proc.stderr:
                    print(bowtie_index_proc.stderr)
                else:
                    self.bowtie_index = Path(
                        f"{str(self.genome_file_abspath.parent / self.genome_name)}"
                    )
                    print(f"Done: {self.bowtie_index }")

            if save_pickle:
                guido_file = Path(
                    f"{str(self.genome_file_abspath.parent / self.genome_name)}.guido"
                )
                with open(guido_file, "wb") as f:
                    pickle.dump(self.__dict__, f, protocol=pickle.HIGHEST_PROTOCOL)
                print(
                    f"{self.genome_name} genome data can now be used by Guido: {str(guido_file)}"
                )

    @property
    def sequence(self):
        if self.is_built:
            return Fasta(str(self.genome_file_abspath))


def load_genome_from_file(guido_file):
    """Load a genome from a .guido file. This file is created when a genome is
    built. It contains all the information needed to use the genome. Guido
    files are saved in the same directory as the genome FASTA file. They are
    named after the genome name and have the .guido extension. Genome object
    can be created by using the :method:`build` method.

    Parameters
    ----------
    guido_file : str
        Path to the .guido file.

    Returns
    -------
    Genome
        :class:`Genome` object.
    """

    if Path(guido_file).exists():
        with open(guido_file, "rb") as f:
            guido_dict = pickle.load(f)
        G = Genome(genome_name=guido_dict["genome_name"])
        for key, val in guido_dict.items():
            setattr(G, key, val)
        return G
    else:
        raise ValueError(f"Provided path to .guido ({guido_file}) file does not exist.")



# flake8: noqa
from .genome import Genome, load_genome_from_file
from .guides import Guide
from .helpers import is_tool, rev_comp
from .locus import Locus, locus_from_coordinates, locus_from_gene, locus_from_sequence



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['guido']

package_data = \
{'': ['*'], 'guido': ['data/*', 'templates/*']}

install_requires = \
['Jinja2>=3.0.1,<4.0.0',
 'h5py>=3.3.0,<4.0.0',
 'mcdm>=1.4,<2.0',
 'ncls>=0.0.60,<0.0.61',
 'numpy>=1.21.1,<2.0.0',
 'pandas>=1.3.1,<2.0.0',
 'pyfaidx>=0.6.1,<0.7.0',
 'pyranges>=0.0.101,<0.0.102',
 'scikit-allel>=1.3.5,<2.0.0',
 'zarr>=2.8.3,<3.0.0']

setup_kwargs = {
    'name': 'guido',
    'version': '0.1.4a0',
    'description': '',
    'long_description': 'None',
    'author': 'Nace Kranjc',
    'author_email': 'nace.kranjc@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>3.9,<=3.10',
}


setup(**setup_kwargs)



import subprocess
import tempfile

import numpy as np

from .helpers import rev_comp


def calculate_ot_sum_score(
    off_targets, ot_mismatch_weights=[10, 5, 4, 3, 1], pam="NGG"
):
    """Calculate sum score for a list of off-targets."""
    return sum(
        [
            ot_mismatch_weights[len(ot["mismatches"]) - pam.count("N")]
            for ot in off_targets
        ]
    )

def get_off_targets_string(off_targets, pam="NGG"):
    """Get a string representation of a list of off-targets."""
    count_matrix = [0, 0, 0, 0, 0]
    
    for ot in off_targets:
        count_matrix[len(ot["mismatches"]) - pam.count("N")] += 1
    
    return '|'.join(
        [str(c) for c in count_matrix]
    )


def calculate_cfd_score(guide, offtargets, mm_scores, pam_scores):
    """Calculate CFD score for a guide and a list of off-targets.

    Adapted from Doench et al. 2016, https://doi.org/10.1038/nbt.3437
    """

    scores = []
    wt = guide.guide_seq[: -len(guide.guide_pam)]

    for ot in offtargets:
        pam = guide.guide_pam[1:]
        sg = ot["seq"][:20]

        score = 1
        sg = sg.replace("T", "U")
        wt = wt.replace("T", "U")
        s_list = list(sg)
        wt_list = list(wt)

        for i, sl in enumerate(s_list):
            if wt_list[i] == sl:
                score *= 1
            else:
                key = (
                    "r" + wt_list[i] + ":d" + rev_comp(sl, rna=True) + "," + str(i + 1)
                )
                score *= mm_scores[key]
        score *= pam_scores[pam]
        scores.append(score)

    return np.array(scores)


def _parse_mismatches(mismatches, strand, seq_len):
    """Parse mismatches from bowtie output.

    Returns a dictionary with mismatches positions as keys and
    mismatched nucleotides as values.
    """
    mm_split = mismatches.split(",")
    mm_dict = {}
    for m in mm_split:
        pos, c = m.split(":")
        c = c.split(">")

        if strand == "-":
            mm_dict[seq_len - int(pos)] = rev_comp(c[0])
        else:
            mm_dict[seq_len - int(pos)] = c[0]
    return mm_dict


def _hit_is_valid(mismatches, non_arbitrary_positions):
    return all(
        [False if pos in non_arbitrary_positions else True for pos in mismatches.keys()]
    )


def run_bowtie(
    guides,
    pam="NGG",
    core_length=10,
    core_mismatches=2,
    total_mismatches=4,
    genome_index_path=None,
    threads=1,
    bowtie_path="bin/bowtie/",
):
    """Run bowtie to find off-targets for a list of guides.

    # TODO: add description of mismatch settings

    Parameters
    ----------
    guides : Guide[]
        List of guides to find off-targets for.
    pam : str, optional
        PAM sequence, by default "NGG"
    core_length : int, optional
        gRNA core length, by default 10
    core_mismatches : int, optional
        Allowed mismatches in core gRNA sequence, by default 2
    total_mismatches : int, optional
        Total allowed mismatches. Max 4, by default 4
    genome_index_path : Genome, optional
        Genome object to search off-targets in, by default None
    threads : int, optional
        Number of threads to run bowtie with, by default 1
    bowtie_path : str, optional
        Path to bowtie binary, by default "bin/bowtie/"

    Returns
    -------
    list
        List of off-targets for each guide.
    """

    pam_mismatches = rev_comp(pam).count("N")
    pam_length = len(pam)
    off_targets = {}

    with tempfile.NamedTemporaryFile(mode="w+t", prefix="guido_") as temp:
        for i, G in enumerate(guides):
            g_seq = rev_comp(G.guide_seq[:-pam_length] + pam)
            temp.write(
                f">{i}|{G.guide_chrom}|{G.guide_start}|{G.guide_end}|{g_seq}|{G.guide_strand}"
            )
            temp.write(f"\n{g_seq}\n")
        temp.seek(0)

        if (core_mismatches + pam_mismatches) > 3:
            raise ValueError(
                f"The value for the parameter core_mismatches is not valid: {core_mismatches}"
            )

        if core_mismatches > total_mismatches:
            raise ValueError(
                f"The value for core_mismatches cannot be greater than total_mismatches: {core_mismatches} > {total_mismatches}"
            )

        bowtie_command = (
            f"{bowtie_path}bowtie -p {threads} --quiet -y "
            f"-n {core_mismatches + pam_mismatches} "
            f"-l {core_length + pam_length} "
            f"-e {(total_mismatches * 30) + 30 } -a "
            f"-x {genome_index_path} "
            f"-f {temp.name}"
        )

        rproc = subprocess.Popen(
            bowtie_command.split(),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        stdout, _ = rproc.communicate()

        non_arbitrary_positions = [
            21 + ix for ix, nucl in enumerate(pam) if nucl in ["A", "T", "G", "C"]
        ]

        # parse bowtie output
        for line in stdout.split("\n"):
            cols = line.split("\t")

            if len(cols) > 1:
                off_target = {}
                ix, g_chrom, g_start, _, g_seq, _ = cols[0].split("|")
                t_strand, t_chrom, t_start, t_seq, _, _, t_mismatches = cols[1:]
                ix = int(ix)

                if ix not in off_targets.keys():
                    off_targets[ix] = []

                if t_strand == "-":
                    t_strand = "+"

                else:
                    t_seq = rev_comp(t_seq)
                    g_seq = rev_comp(g_seq)
                    t_strand = "-"

                mismatches = _parse_mismatches(t_mismatches, t_strand, len(t_seq))

                if (
                    mismatches
                    and _hit_is_valid(mismatches, non_arbitrary_positions)
                    and (g_chrom != t_chrom and int(g_start) != (int(t_start) + 1))
                ):
                    mm_string = "".join(
                        [
                            mismatches[i + 1] if (i + 1) in mismatches.keys() else "."
                            for i, _ in enumerate(g_seq)
                        ]
                    )

                    mm_seq = "".join(
                        [
                            mismatches[i + 1]
                            if (i + 1) in mismatches.keys()
                            else t_seq[i]
                            for i, _ in enumerate(g_seq)
                        ]
                    )

                    off_target["ix"] = ix
                    off_target["mismatches"] = mismatches
                    off_target["mismatches_string"] = mm_string
                    off_target["chromosome"] = t_chrom
                    off_target["start"] = int(t_start)
                    off_target["strand"] = t_strand
                    off_target["seq"] = mm_seq

                    off_targets[ix].append(off_target)

    return off_targets
